from pyspark.sql import SparkSession
from delta.tables import *
import sys

from FabricFramework.FabricLocations import *
from FabricFramework.FabricDataInterface import *
from FabricFramework.FabricConfiguration import *


class FabricFramework:
    """
    Class representing the Fabric Framework for data ingestion.
    """

    def __init__(self, ingestionConfigurationFilePath, ingestionConfigurationTablePath, constants,
                 preProcessFunction=None):
        """
        Initialize a FabricFramework object.

        :param ingestionConfigurationFilePath: Path to the ingestion configuration file.
        :param ingestionConfigurationTablePath: Path to the ingestion configuration table.
        :param constants: Dictionary containing constant values.
        :param preProcessFunction: Optional pre-processing function.
        """
        # Check if the required constants are provided
        self._constantsListChecker(constants)

        self.ingestConfigurationFileLocation = convertURL(ingestionConfigurationFilePath,
                                                          constants['LAKEHOUSE_CONTROL_STAGE_NAME'])
        self.ingestConfigurationTableLocation = convertURL(ingestionConfigurationTablePath,
                                                           constants['LAKEHOUSE_CONTROL_STAGE_NAME'])
        self.preprocess = preProcessFunction
        self.constants = constants
        self.spark = SparkSession.builder.appName("Default_Config").getOrCreate()
        self.fabricInterface = FabricDataInterface()
        self.ingestConfiguration = self.fabricInterface.loadIngestionConfiguration(
            self.ingestConfigurationFileLocation, self.ingestConfigurationTableLocation
        )

    def _constantsListChecker(self, constants):
        """
        Check if the required constants are provided in the constants dictionary.

        :param constants: Dictionary containing constant values.
        :raises SystemExit: If required constants are missing.
        """
        # Ensure constants are provided
        if not constants:
            print("Must supply a constants list that include: workspace, lakehouse and log details.")
            sys.exit(1)
        elif not constants['LAKEHOUSE_CONTROL_STAGE_NAME'] or \
                not constants['LAKEHOUSE_RAW_STAGE_NAME'] or \
                not constants['LAKEHOUSE_TRUSTED_STAGE_NAME'] or \
                not constants['LAKEHOUSE_CURATED_STAGE_NAME']:
            print("Missing Lakehouse details from constants.")
            print("Must supply a constants list that include: workspace, lakehouse and log details.")
            sys.exit(1)

        # Ensure logging level is provided
        if not constants['LOGGING_LEVEL']:
            print("Missing Logging Level from constants.")
            print("Must supply a constants list that include: workspace, lakehouse and log details.")
            sys.exit(1)
        elif constants['LOGGING_LEVEL'] != 'NONE' and \
                (not constants['LOGS_AZURE_TENANT_ID'] or \
                 not constants['LOGS_AZURE_CLIENT_ID'] or \
                 not constants['LOGS_AZURE_CLIENT_SECRET'] or \
                 not constants['DCE_ENDPOINT'] or \
                 not constants['DRC_ID'] or \
                 not constants['LOG_STREAM'] or \
                 not constants['LOG_BUFFER_SIZE']):
            print("Missing Log details from constants.")
            print("Must supply a constants list that include: workspace, lakehouse and log details.")
            sys.exit(1)

    def runIngestion(self, layer=None, system_code=None, sourceID=None):
        """
        Run the ingestion process for the specified layer, system code, and source ID.

        :param layer: Optional layer to ingest (e.g., RAW, TRUSTED).
        :param system_code: Optional system code for the source system.
        :param sourceID: Optional source ID for the source system.
        """
        fabricConfigInstance = FabricConfiguration(
            self.ingestConfiguration, self.constants, self.preprocess,
            listofSourceSystems=system_code, listofSourceID=sourceID
        )

        # Ingest to RAW layer
        if layer is None or self.constants["LAKEHOUSE_RAW_STAGE_NAME"] in layer:
            print(f'Ingesting to {self.constants["LAKEHOUSE_RAW_STAGE_NAME"]}.')
            fabricConfigInstance.loadTables()
            print(f'Ingestion to {self.constants["LAKEHOUSE_RAW_STAGE_NAME"]} Complete.')

        # Load to TRUSTED layer
        if layer is None or self.constants["LAKEHOUSE_TRUSTED_STAGE_NAME"] in layer:
            print(f'Ingesting to {self.constants["LAKEHOUSE_TRUSTED_STAGE_NAME"]}.')
            fabricConfigInstance.validateTables()
            print(f'Ingestion to {self.constants["LAKEHOUSE_TRUSTED_STAGE_NAME"]} Complete.')

        # Flush any remaining logs to the Azure log service
        if self.constants['LOGGING_LEVEL'] != 'NONE':
            print("Saving Logs...", end='')
            fabricConfigInstance.flushLogs()
            print("Logs saved.")
            print("Ingestion Complete.")



from pyspark.sql import SparkSession
from azure.identity import ClientSecretCredential
from azure.keyvault.secrets import SecretClient
from datetime import datetime

from FabricFramework.FabricLocations import *
from FabricFramework.FabricDataInterface import *
from FabricFramework.DataSource import *
from FabricFramework.LogMonitor import *


class FabricConfiguration:
    """
    Configuration class for managing fabric data ingestion and validation processes.
    """

    def __init__(self, ingestConfiguration, constants, function=None, listofSourceSystems=None, listofSourceID=None):
        """
        Initialize the FabricConfiguration.

        Args:
            ingestConfiguration: Configuration for ingestion.
            constants: Dictionary containing constant values.
            function: Optional function to be applied during ingestion.
            listofSourceSystems: Optional list of source systems to filter by.
            listofSourceID: Optional list of source IDs to filter by.
        """
        self.spark = SparkSession.builder.appName("Default_Config").getOrCreate()
        self.fabricInterface = FabricDataInterface()
        self.ingestConfiguration = ingestConfiguration
        self.function = function
        self.constants = constants
        self.sources = self._filterSources(listofSourceSystems, listofSourceID)
        self.logger = Logger.getInstance(
            constants['LOGS_AZURE_TENANT_ID'], constants['LOGS_AZURE_CLIENT_ID'],
            constants['LOGS_AZURE_CLIENT_SECRET'], constants['DCE_ENDPOINT'],
            constants['DRC_ID'], constants['LOG_BUFFER_SIZE']
        )
        self.entity_log_stream = constants['LOG_STREAM']

    def flushLogs(self):
        """
        Flush the logs to the log stream.
        """
        self.logger.flushLogs(self.entity_log_stream)

    def _filterSources(self, listofSourceSystems=None, listofSourceID=None):
        """
        Filter the sources based on the provided source systems and source IDs.

        Args:
            listofSourceSystems: List of source systems to filter by.
            listofSourceID: List of source IDs to filter by.

        Returns:
            Filtered list of sources.
        """
        sources = self.ingestConfiguration.collect()

        if not listofSourceSystems and not listofSourceID:
            return sources

        if listofSourceSystems:
            systems_set = set(listofSourceSystems)
            sources = [source for source in sources if source.system_code in systems_set]

        if listofSourceID:
            sourceid_set = set(listofSourceID)
            sources = [source for source in sources if source.SourceID in sourceid_set]

        return sources

    def validateTables(self, rawDF=None, sourceParams=None):
        """
        Validate tables by processing raw dataframes and saving them to the trusted stage.

        Args:
            rawDF: Raw dataframe to be validated.
            sourceParams: Additional parameters for source.
        """

        # TODO: add the ability to pass in a single raw datagrame and process that into trusted
        if not rawDF:
            for source in self.sources:
                if source["validation_enabled"]:
                    lakehouse_raw = FabricLakehouseLocation(
                        self.constants['LAKEHOUSE_RAW_STAGE_NAME'],
                        source["raw_workspace"],
                        source["raw_lakehouse"]
                    )
                    raw_destination_table = FabricTableLocation(source["destination_name"], lakehouse_raw)
                    df = self.fabricInterface.loadLatestDeltaTable(raw_destination_table)

                    lakehouse_trusted = FabricLakehouseLocation(
                        self.constants['LAKEHOUSE_TRUSTED_STAGE_NAME'],
                        source["trusted_workspace"],
                        source["trusted_lakehouse"]
                    )
                    trusted_destination_table = FabricTableLocation(source["destination_name"], lakehouse_trusted)

                    if self.constants['LOGGING_LEVEL'] != 'NONE':
                        # creates the first initial log in this section
                        assembler = LogAssembler(self.entity_log_stream, self.constants, source, "Processing", self.constants['LAKEHOUSE_TRUSTED_STAGE_NAME'])
                        newLog = assembler.createLog()
                        self.logger.updateAndQueueLog(newLog, None, self.entity_log_stream)

                    # check the save type (CDC or not)
                    if "cdc" not in source["trusted_savetype"].lower():
                        targetCount = self.fabricInterface.saveDeltaTable(
                            sourceDataFrame=df,
                            fabricTableLocation=trusted_destination_table,
                            savetype=source["trusted_savetype"],
                            primarykey=source["primary_key"]
                        )
                    else: #TODO: create a CDC function call
                        pass

                    # now log the save event.
                    if self.constants['LOGGING_LEVEL'] != 'NONE':
                        updated_log = {
                            "EndDateTime": datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"),
                            "LogDateTime": datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"),
                            "Status": "Completed", "SourceRows": df.count(), "TargetRows": targetCount
                        }
                        log = Log(self.entity_log_stream, updated_log)
                        self.logger.updateAndQueueLog(newLog, log, self.entity_log_stream)

                else: # if the validation is not enabled for the SourceID
                    print(f'Skipping SourceID: {source["SourceID"]} as it is disabled in the ingestion configuration.')
                    if self.constants['LOGGING_LEVEL'] in ['INFO', 'DEBUG']:
                        updated_log = {
                            "EndDateTime": datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"),
                            "LogDateTime": datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"),
                            "Status": "Skipped"
                        }
                        log = Log(self.entity_log_stream, updated_log)
                        self.logger.updateAndQueueLog(newLog, log, self.entity_log_stream)



    # TODO: Option to load only a single table or multiple (provided as a list)
    def loadTables(self):
        """
        Load tables from the sources, preprocess and save them to the raw stage.
        """
        for source in self.sources:
            print(f'Evaluating SourceID: {source["SourceID"]}')

            if source["ingestion_enabled"]:
                print(f'Ingestion enabled for SourceID: {source["SourceID"]} and ingestion started.')

                if source["source_type"] == 'FABRIC-TEXT':
                    connection = source["raw_workspace"] + ";" + source["raw_lakehouse"]
                elif source["source_type"] == 'FABRIC-TABLE':
                    connection = source["prelanding_workspace"] + ";" + source["prelanding_lakehouse"]
                else:
                    keyvaultConnection = KeyVault(
                        source["key_vault_name"], source["tenant_id"], source["client_id"],
                        source["client_secret"], source["secret_name"]
                    )
                    connection = keyvaultConnection.secretValue

                newDataSource = DataSource.getDataSourceType(connection, source, self.function)
                loaded_data = newDataSource.loadTable()

                lakehouse_raw = FabricLakehouseLocation(
                    self.constants['LAKEHOUSE_RAW_STAGE_NAME'],
                    source["raw_workspace"],
                    source["raw_lakehouse"]
                )
                raw_destination_table = FabricTableLocation(source["destination_name"], lakehouse_raw)

                if self.constants['LOGGING_LEVEL'] != 'NONE':
                    # creates the first initial log
                    assembler = LogAssembler(self.entity_log_stream, self.constants, source, "Processing", self.constants['LAKEHOUSE_RAW_STAGE_NAME'])
                    newLog = assembler.createLog()
                    self.logger.updateAndQueueLog(newLog, None, self.entity_log_stream)

                targetCount = self.fabricInterface.saveDeltaTable(
                    loaded_data,
                    raw_destination_table,
                    savetype=source["raw_savetype"],
                    primarykey=source["primary_key"]
                )

                if self.constants['LOGGING_LEVEL'] != 'NONE':
                    updated_log = {
                        "EndDateTime": datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"),
                        "LogDateTime": datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"),
                        "Status": "Completed", "SourceRows": loaded_data.count(), "TargetRows": targetCount
                    }
                    log = Log(self.entity_log_stream, updated_log)
                    self.logger.updateAndQueueLog(newLog, log, self.entity_log_stream)

            else:
                print(f'Skipping SourceID: {source["SourceID"]} as it is disabled in the ingestion configuration.')
                if self.constants['LOGGING_LEVEL'] in ['INFO', 'DEBUG']:
                    updated_log = {
                        "EndDateTime": datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"),
                        "LogDateTime": datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"),
                        "Status": "Skipped"
                    }
                    log = Log(self.entity_log_stream, updated_log)
                    self.logger.updateAndQueueLog(newLog, log, self.entity_log_stream)


class KeyVault:
    """
    Class for interacting with Azure Key Vault to retrieve secrets.
    """

    def __init__(self, keyvaultname, tenantid, clientid, clientsecret, secretName):
        """
        Initialize the KeyVault client.

        Args:
            keyvaultname: Name of the Key Vault.
            tenantid: Azure Tenant ID.
            clientid: Azure Client ID.
            clientsecret: Azure Client Secret.
            secretName: Name of the secret to retrieve.
        """
        credential = ClientSecretCredential(tenantid, clientid, clientsecret)
        vaultURL = f"https://{keyvaultname}.vault.azure.net"
        secretClient = SecretClient(vault_url=vaultURL, credential=credential)
        self.secretVal = secretClient.get_secret(secretName).value

    @property
    def secretValue(self):
        """
        Retrieve the secret value.

        Returns:
            Secret value from the Key Vault.
        """
        return self.secretVal



# libraries
from pyspark.sql import SparkSession, Row
from pyspark.sql.types import StructType, StructField, StringType, IntegerType, TimestampType, BooleanType, ArrayType
from pyspark.sql.functions import lit, when, col, current_timestamp, monotonically_increasing_id, from_json, explode
import pyspark.sql.functions as F
import json
import re
import great_expectations as gx
from great_expectations.checkpoint import Checkpoint
from delta.tables import *
import time

from concurrent.futures import ThreadPoolExecutor
from multiprocessing.pool import ThreadPool

from FabricFramework.FabricDataInterface import *

""" Constants that would be utilized within the Great Expectation Implementation """

# Below constants represent the types of validation
IS_NULL = "expect_column_values_to_not_be_null"
IS_OUT_OF_RANGE = "expect_column_values_to_be_between" # The first value is the minimum, whilst the second is the maximum
IS_NOT_MATCHING_REGEX = "expect_column_values_to_match_regex"
IS_NOT_EXPECTED_DATA_TYPE = "expect_column_values_to_be_of_data_type"

# Below constants represent the Data Type validations
INTEGER_TYPE = "Integer"
DOUBLE_TYPE = "Double"
BOOLEAN_TYPE = "Boolean"


class Expectation:
    """ The Expectation class represent the information provided for by the user for a validation that needs to be applied to a column. """

    def __init__(self, name, column, _pass, options=[]):
        self.name = name
        self.column = column
        self._pass = _pass
        self.options = options if name != IS_NOT_EXPECTED_DATA_TYPE else self.generateDataTypeValidationRegex(options)

    def generateDataTypeValidationRegex(self, options):
        """ When it comes to a Data Type validation, this method is used to create a Regex to validate if the values match the provided Data Type for the validation. """
        if (options[0] == INTEGER_TYPE):
            options.append("^-?\\d+$")
        elif (options[0] == DOUBLE_TYPE):
            options.append("^-?\\d+(\\.\\d+)?$")
        elif (options[0] == BOOLEAN_TYPE):
            options.append("^(0|1|true|false|True|False|TRUE|FALSE)$")
        return options


class GreatExpectationsValidation:

    def __init__(self, source_table_name, df, trusted_workspace_id, trusted_lakehouse_id, list_validation,
                 trusted_metadata=None, trusted_savetype='overwrite', sk=None, datasource_name="my_datasource_name",
                 data_asset_name="my_data_asset_name", expectation_suite_name="my_expectation_suite_name",
                 checkpoint_name="my_checkpoint"):
        self.df = df
        self.sk = sk if sk is not None else self.generateIndexColumn()  # if this is null -> Generate a Index column to the df
        self.source_table_name = source_table_name
        self.trusted_workspace_id = trusted_workspace_id
        self.trusted_lakehouse_id = trusted_lakehouse_id
        self.trusted_savetype = trusted_savetype
        self.list_validation = list_validation
        self.trusted_metadata = trusted_metadata
        self.datasource_name = datasource_name
        self.data_asset_name = data_asset_name
        self.expectation_suite_name = expectation_suite_name
        self.checkpoint_name = checkpoint_name
        self.context = gx.get_context()
        self.checkpoint_result = None
        self.invalid_row_information_df = None
        self.df = self.df.withColumn("is_valid", lit(True))
        self.df = self.df.withColumn("pass", lit(True))
        self.fabricInterface = FabricDataInterface()
        self.spark = SparkSession.builder.appName("Default_Config").getOrCreate()

    def generateIndexColumn(self):
        """ This method is used to add an Index/SK column if the dataset doesnt have a SK column """
        self.df = self.df.withColumn("Index_SK", monotonically_increasing_id() + 1)
        return "Index_SK"

    def validateData(self):
        """ This method utilizes the information initialized within the _init_ and helps to validate the data whilst returning the result in a Dictionary format  """
        datasource = self.context.sources.add_spark(self.datasource_name)
        data_asset = datasource.add_dataframe_asset(name=self.data_asset_name)
        my_batch_request = data_asset.build_batch_request(dataframe=self.df)

        # Creates an Expecation suite, which is a collection of verifiable assertions about data
        self.context.add_or_update_expectation_suite(self.expectation_suite_name)
        validator = self.context.get_validator(batch_request=my_batch_request,
                                               expectation_suite_name=self.expectation_suite_name)

        # ucrrent issue
        validator = self.setExpectations(validator)

        validator.save_expectation_suite(discard_failed_expectations=False)

        # Creates a Checkpoint, which is the primary means for validating data in a production deployment of Great Expectations
        checkpoint = Checkpoint(
            name=self.checkpoint_name,
            run_name_template="%Y%m%d-%H%M%S-my-run-name-template",
            data_context=self.context,
            batch_request=my_batch_request,
            expectation_suite_name=self.expectation_suite_name,
            action_list=[
                {
                    "name": "store_validation_result",
                    "action": {"class_name": "StoreValidationResultAction"},
                },
                {
                    "name": "store_evaluation_params",
                    "action": {"class_name": "StoreEvaluationParametersAction"},
                },
                {"name": "update_data_docs", "action": {"class_name": "UpdateDataDocsAction"}},
            ],
            runtime_configuration={
                "result_format": {
                    "result_format": "COMPLETE",
                    "unexpected_index_column_names": [self.sk],
                    "return_unexpected_index_query": True,
                    "include_unexpected_rows": False
                },
            },
        )
        self.context.add_or_update_checkpoint(checkpoint=checkpoint)
        self.checkpoint_result = checkpoint.run()
        # self._generateInvalidRowInformationTable()
        self._generateValidAndInvalidRowTables()

        return self.checkpoint_result

    def buildDocs(self):
        """ The method helps to build and save your validation result informaiton as a Data Doc """
        self.context.build_data_docs()
        self.context.open_data_docs()

    def setExpectations(self, validator):
        """ The method helps to set each type of validation to the validator (Expecation Suite) by assessing the user provided validation type """

        for validation in self.list_validation:

            print(validation.column)
            print(validation.name)

            if validation.name == IS_NULL:
                validator.expect_column_values_to_not_be_null(
                    column=validation.column
                    , meta={
                        "key": validation.column + "_" + validation.name
                        , "pass": validation._pass
                        , "display_name": "Null Validation: " + validation.column
                    }
                )
            elif validation.name == IS_OUT_OF_RANGE:
                validator.expect_column_values_to_be_between(
                    column=validation.column
                    , min_value=validation.options[0]
                    , max_value=validation.options[1]
                    , meta={
                        "key": validation.column + "_" + validation.name
                        , "pass": validation._pass
                        , "display_name": "Range Validation: " + validation.column
                    }
                )
            elif validation.name == IS_NOT_MATCHING_REGEX:
                validator.expect_column_values_to_match_regex(
                    column=validation.column
                    , regex=validation.options[0]
                    , meta={
                        "key": validation.column + "_" + validation.name + "_" + str(validation.options[0])
                        , "pass": validation._pass
                        , "display_name": "Regex Validation: " + validation.column
                    }
                )
            elif validation.name == IS_NOT_EXPECTED_DATA_TYPE:
                validator.expect_column_values_to_match_regex(
                    column=validation.column
                    , regex=validation.options[1]
                    , meta={
                        "key": validation.column + "_" + validation.name + "_" + validation.options[0]
                        , "pass": validation._pass
                        , "display_name": validation.options[0] + " Validation: " + validation.column
                    }
                )

        return validator

    # Method is used to create the Invalid Row Information table - stores the SK, validation type, and the column of which the validation was applied.
    def _generateInvalidRowInformationTable(self):
        """ The method reads the Validation Result (Checkpoint Result) and creates/overwrites the Invalid Row Information Table """
        pattern = r'^ValidationResultIdentifier'
        validationResultIdentifierProperty = next(
            (key for key in self.checkpoint_result["run_results"] if re.search(pattern, str(key))), None)

        if (
        not self.checkpoint_result["run_results"][validationResultIdentifierProperty]["validation_result"]["success"]):
            schema = StructType([
                StructField(self.sk, IntegerType(), True),
                StructField("column_name", StringType(), True),
                StructField("validation_type", StringType(), True),
                StructField("display_name", StringType(), True),
                StructField("timestamp", TimestampType(), True),
                StructField("pass", BooleanType(), True)
            ])
            self.invalid_row_information_df = self.spark.createDataFrame([], schema=schema)
            for result in \
            self.checkpoint_result["run_results"][validationResultIdentifierProperty]["validation_result"]["results"]:
                if (not result["success"]):
                    filtered_invalid_rows = self.spark.createDataFrame([], schema=schema)
                    queried_rows = eval(self.formatQueryString(("self." + result["result"]["unexpected_index_query"]),
                                                               result["expectation_config"]["expectation_type"]))
                    filtered_invalid_rows = queried_rows.select(col(self.sk).cast("int"))
                    filtered_invalid_rows = filtered_invalid_rows.withColumn("column_name", lit(
                        result["expectation_config"]["kwargs"]["column"]))
                    filtered_invalid_rows = filtered_invalid_rows.withColumn("validation_type", lit(
                        result["expectation_config"]["meta"]["key"]))
                    filtered_invalid_rows = filtered_invalid_rows.withColumn("display_name", lit(
                        result["expectation_config"]["meta"]["display_name"]))
                    filtered_invalid_rows = filtered_invalid_rows.withColumn("timestamp", lit(current_timestamp()))

                    # Update the is_valid column
                    invalid_sks = [int(row[self.sk]) for row in filtered_invalid_rows.collect()]
                    self.df = self.df.withColumn("is_valid",
                                                 when(col(self.sk).cast("int").isin(invalid_sks), False).otherwise(
                                                     col("is_valid")))  # set is_valid values to the main df

                    # Update the pass column: If the _pass value is False AND the validation is False -> Set Pass to False
                    if (not result["expectation_config"]["meta"]["pass"]):
                        filtered_invalid_rows = filtered_invalid_rows.withColumn("pass", lit(False))
                        self.df = self.df.withColumn("pass", when(col(self.sk).cast("int").isin(invalid_sks),
                                                                  lit(False)).otherwise(
                            col("pass")))  # set pass values to the main df
                    else:
                        filtered_invalid_rows = filtered_invalid_rows.withColumn("pass", lit(True))

                    # Union the filtered invalid rows to the dataframe storing all the invalid rows
                    self.invalid_row_information_df = self.invalid_row_information_df.union(filtered_invalid_rows)

            # Creates/Overwrites the Invalid row information table
            # loadToTrusted(self.invalid_row_information_df, 0, f"abfss://{self.trusted_workspace_id}@onelake.dfs.fabric.microsoft.com/{self.trusted_lakehouse_id}/Tables/{self.source_table_name}_invalid_row_information")
            print("Writing invalid row information table...", end="")
            self.fabricInterface.writeDeltaTable(self.trusted_workspace_id, self.trusted_lakehouse_id,
                                f'{self.source_table_name}_invalid_row_information', self.invalid_row_information_df,
                                self.trusted_savetype)
            print("DONE.")

    # Method is used to generate the Fabric table of both valid and invalid rows of the df.
    def _generateValidAndInvalidRowTables(self):
        """ The method reads through the updated df and filters out the df to Invalid and Valid data, which then helps to create the Invalid and Valid Tables """
        # Drop the 'validtion_type' and the 'column_name' columns
        self.df = self.df.drop('validation_type', 'column_name')

        # Create source_table_meta_information Fabric table
        schema = StructType([
            StructField("source_table", StringType(), True),
            StructField("row_count", IntegerType(), True)
        ])
        meta_info_df = self.spark.createDataFrame([(self.source_table_name, self.df.count())], schema=schema)

        # print("Writing meta information table...", end="")
        # fdi.writeDeltaTable(self.trusted_workspace_id, self.trusted_lakehouse_id, f'{self.source_table_name}_meta_info', meta_info_df, self.trusted_savetype)
        # loadToTrusted(meta_info_df, 0, f"abfss://{trusted_workspace_id}@onelake.dfs.fabric.microsoft.com/{self.trusted_lakehouse_id}/Tables/{self.source_table_name}_meta_info")
        # print("DONE.")

        # Filter the valid rows, proceed with data type conversion and finally create/overwrite the Valid rows table
        valid_rows_df = self.df.where(
            (col("is_valid") == True) | ((col("is_valid") == False) & (col("pass") == True))).drop('is_valid')
        # valid_rows_df = self.convertToDataType(valid_rows_df)

        print("Writing valid rows table...", end="")

        print(f'second test: sk is {self.sk}')

        if (self.trusted_savetype == "masters" or self.trusted_savetype == "incremental"):
            print("calling cdc method, save type: ")
            self.fabricInterface.writeCDCTable(self.trusted_workspace_id, self.trusted_lakehouse_id, self.source_table_name,
                              valid_rows_df, savetype=self.trusted_savetype, primarykey=self.sk)
        else:
            print("calling write delta table")
            self.fabricInterface.writeDeltaTable(self.trusted_workspace_id, self.trusted_lakehouse_id, self.source_table_name,
                                valid_rows_df, savetype=self.trusted_savetype)
            # loadToTrusted(valid_rows_df, self.incremental_load, f"abfss://{trusted_workspace_id}@onelake.dfs.fabric.microsoft.com/{self.trusted_lakehouse_id}/Tables/{self.source_table_name}", self.sk)

        print("DONE.")

        # Filter the invalid rows and create/overwrite the Invalid rows table
        invalid_rows_df = self.df.where(col("is_valid") == False).drop('is_valid')

        # print("Writing invalid rows table...", end="")
        # fdi.writeDeltaTable(self.trusted_workspace_id, self.trusted_lakehouse_id, f'{self.source_table_name}_invalid_rows', invalid_rows_df, self.trusted_savetype)
        # loadToTrusted(valid_rows_df, self.incremental_load, f"abfss://{trusted_workspace_id}@onelake.dfs.fabric.microsoft.com/{self.trusted_lakehouse_id}/Tables/{self.source_table_name}_invalid_rows", self.sk)
        print("DONE.")

    def formatQueryString(self, query_string, expectation_name):
        """ The method helps to wrap the Expression of the query string in double quotes"""
        query_string = (query_string[:22] + "\"" + query_string[22:-2] + "\"" + query_string[-2:])
        if (expectation_name == "expect_column_values_to_match_regex"):
            query_string = self.wrapSingleQuotes(query_string)
        return query_string

    def wrapSingleQuotes(self, query_string):
        """ The method helps to wrap the Regex expression of the query string in single quotes"""
        rlike_section = query_string.split('RLIKE(')[1]
        remove_ending_brackets = rlike_section[:-5]
        final_string = query_string.split('RLIKE(')[0] + "RLIKE(" + remove_ending_brackets.split(', ', 2)[
            0] + "," + "\'" + remove_ending_brackets.split(', ', 2)[1] + "\'" + rlike_section[-5:]
        return final_string

    def convertToDataType(self, valid_rows_df):
        """ The method helps to convert the successfuly validated columns to their respective valid data type for Trusted Loading """
        for column, data_type in self.trusted_metadata:
            if data_type == "string":
                valid_rows_df = valid_rows_df.withColumn(column, col(column).cast("string"))
            elif data_type == "int":
                valid_rows_df = valid_rows_df.withColumn(column, col(column).cast("int"))
            elif data_type == "double":
                valid_rows_df = valid_rows_df.withColumn(column, col(column).cast("double"))
            elif data_type == "boolean":
                valid_rows_df = valid_rows_df.withColumn(column, col(column).cast("boolean"))
        return valid_rows_df


from pyspark.sql import SparkSession
from FabricFramework.FabricLocations import *
from FabricFramework.FabricDataInterface import *

class DataSource(object):
    """Base class for data sources using Spark sessions."""

    def __init__(self):
        """Initialize Spark session with a default application name."""
        self.spark = SparkSession.builder.appName("Default").getOrCreate()
        self.fabricInterface = FabricDataInterface()

    @staticmethod
    def getDataSourceType(connection, source, function=None):
        """Factory method to get the appropriate data source type based on the source type."""
        sourceType = source["source_type"].upper()

        if sourceType == 'SQL':
            return SQLDataSource(connection, source, function)
        elif sourceType == 'BLOB-CSV':
            return BLOBCSVDataSource(connection, source, function)
        elif sourceType == 'BLOB-PARQUET':
            return BLOBParquetDataSource(connection, source, function)
        elif sourceType == 'FABRIC-TEXT':
            return FabricTextDataSource(connection, source, function)
        elif sourceType == 'FABRIC-TABLE':
            return FabricTableDataSource(connection, source, function)

    def loadTable(self):
        """Placeholder method to be overridden in subclass."""
        pass

    def execute_function(self, function, *args, **kwargs):
        """Execute a function with the provided arguments and keyword arguments."""
        return function(*args, **kwargs)

class SQLDataSource(DataSource):
    """Data source for SQL databases."""

    def __init__(self, connection, source, function=None):
        """Initialize SQL data source with connection details and table."""
        self.url = connection
        self.driver = "com.microsoft.sqlserver.jdbc.SQLServerDriver"  # JDBC driver for SQL Server
        self.source = source
        self.source_name = source["source_name"]
        self.preProcessingEnabled = source["preprocessing_enabled"]
        self.p_options = source["preprocessing_options"]
        self.function = function
        super().__init__()

    def loadTable(self):
        """Load table using JDBC with the specified connection properties."""
        connection_properties = {
            "driver": self.driver,
            "url": self.url
        }

        if self.preProcessingEnabled:
            self.execute_function(self.function, self.source)

        df = self.spark.read.jdbc(url=self.url, table=self.source_name, properties=connection_properties)
        return df

class BLOBCSVDataSource(DataSource):
    """Data source for loading CSV files from Azure Blob storage."""

    def __init__(self, connection, source, function=None):
        """Initialize data source with connection details, table (blob path), header, and schema."""
        self.url = connection
        self.source_name = source["source_name"]
        self.header = source["header"]
        self.schema = source["schema"]
        super().__init__()

    def loadTable(self):
        """Load a CSV file from Azure Blob using Spark."""
        connection_tokens = self.url.split(';')
        storage_account_name = connection_tokens[1].split('=')[1]
        storage_account_access_key = connection_tokens[2].split('=', 1)[1]
        blob_container = connection_tokens[4].split('=')[1]

        self.spark.conf.set(f'fs.azure.account.key.{storage_account_name}.blob.core.windows.net', storage_account_access_key)

        blob_relative_path = self.source_name
        blob_url = f"wasbs://{blob_container}@{storage_account_name}.blob.core.windows.net/{blob_relative_path}"
        df = self.spark.read.format("csv").option("header", self.header).load(blob_url)
        return df

class BLOBParquetDataSource(DataSource):
    """Data source for loading Parquet files from Azure Blob storage."""

    def __init__(self, connection, source, function=None):
        """Initialize data source with connection details, table (blob path), header, and schema."""
        self.url = connection
        self.source_name = source("source_name")
        self.header = source("header")
        self.schema = source("schema")
        super().__init__()

    def loadTable(self):
        """Load a Parquet file from Azure Blob using Spark."""
        connection_tokens = self.url.split(';')
        storage_account_name = connection_tokens[1].split('=')[1]
        storage_account_access_key = connection_tokens[2].split('=', 1)[1]
        blob_container = connection_tokens[4].split('=')[1]

        self.spark.conf.set(f'fs.azure.account.key.{storage_account_name}.blob.core.windows.net', storage_account_access_key)

        blob_relative_path = self.source_name
        blob_url = f"wasbs://{blob_container}@{storage_account_name}.blob.core.windows.net/{blob_relative_path}"
        if self.schema:
            df = self.spark.read.schema(self.schema).parquet(blob_url)
        else:
            df = self.spark.read.parquet(blob_url)
        return df

class FabricTextDataSource(DataSource):
    """Data source for loading text files from a Fabric container."""

    def __init__(self, connection, source, function=None):
        """Initialize data source with connection details and file-specific options."""
        self.url = connection
        self.file = source["source_name"]
        self.header = source["header"]
        self.schema = source["schema"]
        self.destination = source["destination_name"]
        self.preProcessingEnabled = source["preprocessing_enabled"]
        self.p_options = source["preprocessing_options"]
        self.function = function
        self.source = source
        super().__init__()

    def loadTable(self):
        """Load a text file from Fabric using configured options and execute any preprocessing functions if specified."""
        if self.url:
            connection_tokens = self.url.split(';')
            workspace_id = connection_tokens[0]
            lakehouse_id = connection_tokens[1]

            # Create lakehouse location
            lakehouse_datasource = FabricLakehouseLocation("FabricTextSource", workspace_id, lakehouse_id)

            if self.preProcessingEnabled:
                return self.execute_function(self.function, self.source, self.file, lakehouse_datasource, self.destination, self.p_options, self.header)

        return None

class FabricTableDataSource(DataSource):
    def __init__(self, connection, source, function=None):
        """Initialize data source with connection details and file-specific options."""
        self.url = connection
        self.file = source("source_name")
        self.header = source("header")
        self.schema = source("schema")
        self.destination = source["destination_name"]
        self.preProcessingEnabled = source["preprocessing_enabled"]
        self.p_options = source["preprocessing_options"]
        self.function = function
        self.source = source
        super().__init__()

    def loadTable(self):
        """Load a text file from Fabric using configured options and execute any preprocessing functions if specified."""
        if self.url:
            connection_tokens = self.url.split(';')
            workspace_id = connection_tokens[0]
            lakehouse_id = connection_tokens[1]

            # Create lakehouse location
            lakehouse_datasource = FabricLakehouseLocation("FabricTableSource", workspace_id, lakehouse_id)

            if self.preProcessingEnabled:
                return self.execute_function(self.function, self.source, lakehouse_datasource)
            else:
                # create the table location
                tableLocation = FabricTableLocation(self.source['source_name'], lakehouse_datasource)
                # return the dataframe as is (without preprocessing)
                return self.fabricInterface.loadLatestDeltaTable(tableLocation)

        print("Cannot load table, no table location provided.")
        return None



from pyspark.sql import SparkSession


# TODO: import this from the constants list defined in an external notebook
# Define a list to hold lakehouse stage names
LAKEHOUSE_STAGE_LIST = []

# Define constants for different lakehouse stages
LAKEHOUSE_CONTROL_STAGE_NAME = 'CONTROL'
LAKEHOUSE_PREPROCESSING_STAGE_NAME = 'PREPROCESSING'
LAKEHOUSE_RAW_STAGE_NAME = 'RAW'
LAKEHOUSE_TRUSTED_STAGE_NAME = 'TRUSTED'
LAKEHOUSE_CURATED_STAGE_NAME = 'CURATED'

# Extend the stage list with the defined stage names
LAKEHOUSE_STAGE_LIST.extend([
    LAKEHOUSE_CONTROL_STAGE_NAME,
    LAKEHOUSE_PREPROCESSING_STAGE_NAME,
    LAKEHOUSE_RAW_STAGE_NAME,
    LAKEHOUSE_TRUSTED_STAGE_NAME,
    LAKEHOUSE_CURATED_STAGE_NAME
])


class FabricLakehouseLocation:
    """
    Class representing a Fabric Lakehouse Location.
    """

    def __init__(self, stage, workspace, lakehouse):
        """
        Initialize a FabricLakehouseLocation object.

        :param stage: Stage of the lakehouse.
        :param workspace: Workspace ID.
        :param lakehouse: Lakehouse ID.
        """
        self._stage = stage
        self._workspace = workspace
        self._lakehouse = lakehouse
        self.spark = SparkSession.builder.appName("Default_Config").getOrCreate()

    @property
    def stage(self):
        return self._stage

    @property
    def workspace(self):
        return self._workspace

    @property
    def lakehouse(self):
        return self._lakehouse

    def abfss_path(self, path=None):
        """
        Generate an ABFSS path for the lakehouse location.

        :param path: Optional path within the lakehouse.
        :return: ABFSS path as a string.
        """
        if path:
            return f'abfss://{self._workspace}@onelake.dfs.fabric.microsoft.com/{self._lakehouse}.Lakehouse/{path}'
        return f'abfss://{self._workspace}@onelake.dfs.fabric.microsoft.com/{self._lakehouse}.Lakehouse'

    @stage.setter
    def stage(self, stage):
        """
        Setter for the stage property, with validation against the predefined stage list.

        :param stage: Stage to be set.
        :raises ValueError: If the stage is not in the predefined stage list.
        """
        if stage not in LAKEHOUSE_STAGE_LIST:
            raise ValueError("Stage name must be defined in the LAKEHOUSE_STAGE_LIST")
        self._stage = stage


class FabricLocation:
    """
    Base class for Fabric locations.
    """

    def abfss_path(self):
        pass


class FabricFileLocation(FabricLocation):
    """
    Class representing a file location in the lakehouse.
    """

    def __init__(self, file, lakehouseLocation=None):
        """
        Initialize a FabricFileLocation object.

        :param file: File name.
        :param lakehouseLocation: Optional FabricLakehouseLocation object.
        """
        if lakehouseLocation:
            self.location = lakehouseLocation.abfss_path(f'Files/{file}')
        else:
            self.location = f'Files/{file}'

    def abfss_path(self):
        return self.location


class FabricFolderLocation(FabricLocation):
    """
    Class representing a folder location in the lakehouse.
    """

    def __init__(self, folder, lakehouseLocation=None):
        """
        Initialize a FabricFolderLocation object.

        :param folder: Folder name.
        :param lakehouseLocation: Optional FabricLakehouseLocation object.
        """
        if lakehouseLocation:
            self.location = lakehouseLocation.abfss_path(f'Files/{folder}')
        else:
            self.location = f'Files/{folder}'

    def abfss_path(self):
        return self.location


class FabricTableLocation(FabricLocation):
    """
    Class representing a table location in the lakehouse.
    """

    def __init__(self, table, lakehouseLocation=None):
        """
        Initialize a FabricTableLocation object.

        :param table: Table name.
        :param lakehouseLocation: Optional FabricLakehouseLocation object.
        """
        if lakehouseLocation:
            self.location = lakehouseLocation.abfss_path(f'Tables/{table}')
        else:
            self.location = f'Tables/{table}'

    def abfss_path(self):
        return self.location


def extract_url(url):
    """
    Extract components from a given ABFSS URL.

    :param url: ABFSS URL as a string.
    :return: Tuple containing workspace ID, lakehouse ID, entity name, and entity type.
    :raises ValueError: If the URL does not contain 'Tables/' or 'Files/'.
    """
    core_part = url.split('://')[1]

    # Extract the workspace_id
    workspace_id, rest = core_part.split('@onelake.dfs.fabric.microsoft.com/')

    # Extract the lakehouse_id and the latter part which could contain Tables or Files
    lakehouse_id, latter_part = rest.split('.Lakehouse/')

    # Check if the latter part contains 'Tables' or 'Files' and extract table_name or file_name
    if 'Tables/' in latter_part:
        entity_type, entity_name = latter_part.split('/')
    elif 'Files/' in latter_part:
        entity_type, entity_name = latter_part.split('/')
    else:
        raise ValueError("URL should contain either 'Tables/' or 'Files/'")

    return workspace_id, lakehouse_id, entity_name, entity_type


def convertURL(url, lakehouseName):
    """
    Convert a given URL to a FabricLocation object.

    :param url: ABFSS URL as a string.
    :param lakehouseName: Name of the lakehouse.
    :return: FabricLocation object representing the converted URL.
    :raises ValueError: If the URL does not contain 'Tables/' or 'Files/'.
    """
    workspace_id, lakehouse_id, entity_name, type = extract_url(url)

    # Create lakehouse location
    lakehouse = FabricLakehouseLocation(lakehouseName, workspace_id, lakehouse_id)

    # Create file or table location
    if type == 'Files':
        location = FabricFileLocation(entity_name, lakehouse)
    elif type == 'Tables':
        location = FabricTableLocation(entity_name, lakehouse)
    else:
        raise ValueError("URL should contain either 'Tables/' or 'Files/'")

    return location



import os
from azure.identity import DefaultAzureCredential
from azure.monitor.ingestion import LogsIngestionClient
from azure.core.exceptions import HttpResponseError
from datetime import datetime
import threading


class Log:
    """
    Class representing a log entry with a stream name and log information.
    """
    def __init__(self, stream_name, log_info):
        """
        Initialize a Log object.

        :param stream_name: Name of the log stream.
        :param log_info: Dictionary containing log information.
        """
        self._stream_name = stream_name
        self._log_info = log_info

    @property
    def log_info(self):
        return self._log_info

    @property
    def stream_name(self):
        return self._stream_name


class LogAssembler:
    """
    Class for assembling log entries with given parameters.
    """

    def __init__(self, stream_name, constants, source, status, layer):
        """
        Initialize a LogAssembler object.

        :param stream_name: Name of the log stream.
        :param constants: Dictionary containing constant values for the log.
        :param source: Dictionary containing source information.
        :param status: Status of the log entry.
        :param layer: Layer information for the log entry.
        """
        self.stream_name = stream_name
        self.constants = constants
        self.source = source
        self.entity_log = {
            "LogRunID": self.constants['LOGRUN_ID'],
            "UserName": self.constants['USERNAME'],
            "Group": source["system_code"],
            "SourceID": source["SourceID"],
            "Entity": source["source_name"],
            "Layer": layer,
            "Source": source["source_type"],
            "Target": source["destination_name"],
            "StartDateTime": self.constants['EXECUTION_STARTDATETIME'],
            "EndDateTime": datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"),
            "LogDateTime": datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"),
            "LogType": self.constants['LOGGING_LEVEL'],
            "Status": status,
            "SourceRows": "",
            "TargetRows": "",
            "Error": "",
            "ValidationRules": "",
            "Parameters": source["preprocessing_options"]
        }

    def createLog(self):
        return Log(self.stream_name, self.entity_log)


class Logger:
    """
    Singleton class for managing and uploading logs to Azure.
    """
    _instance = None

    @classmethod
    def getInstance(cls, *args, **kwargs):
        """
        Get the singleton instance of Logger.

        :param args: Positional arguments.
        :param kwargs: Keyword arguments.
        :return: Singleton instance of Logger.
        """
        if not cls._instance:
            cls._instance = cls(*args, **kwargs)
        return cls._instance

    def __init__(self, az_tenant_id, az_client_id, az_client_secret, dce_endpoint, dcr_immutableid, buffer=1000):
        """
        Initialize the Logger object with Azure credentials and configuration.

        :param az_tenant_id: Azure tenant ID.
        :param az_client_id: Azure client ID.
        :param az_client_secret: Azure client secret.
        :param dce_endpoint: Data Collection Endpoint URL.
        :param dcr_immutableid: Data Collection Rule immutable ID.
        :param buffer: Buffer size for log batching.
        """
        if not hasattr(self, 'initialised'):  # This checks if the __init__ has been called before
            self.initialised = True
            self.logs = []
            self.az_tenant_id = az_tenant_id
            self.az_client_id = az_client_id
            self.az_client_secret = az_client_secret
            self.dce_endpoint = dce_endpoint
            self.dcr_immutableid = dcr_immutableid
            self.buffer = buffer

            # Set Azure credentials as environment variables
            os.environ["AZURE_TENANT_ID"] = self.az_tenant_id
            os.environ["AZURE_CLIENT_ID"] = self.az_client_id
            os.environ["AZURE_CLIENT_SECRET"] = self.az_client_secret

    def appendLogs(self, log_entry):
        self.logs.append(log_entry)

    def saveLogs(self, log_batch, stream_name):
        """
        Save a batch of logs to Azure Monitor.

        :param log_batch: List of log entries.
        :param stream_name: Name of the log stream.
        """
        credential = DefaultAzureCredential()
        client = LogsIngestionClient(endpoint=self.dce_endpoint, credential=credential, logging_enable=True)

        try:
            client.upload(rule_id=self.dcr_immutableid, stream_name=stream_name, logs=log_batch)
        except HttpResponseError as e:
            print(f"Log upload failed: {e}")

    def queueLog(self, log):
        """
        Queue a log entry for uploading. If the buffer is full, upload the current batch of logs.

        :param log: Log object to be queued.
        """
        if len(self.logs) < self.buffer:
            self.appendLogs(log.log_info)
        else:
            log_batch = self.logs
            thread = threading.Thread(target=self.saveLogs, args=(log_batch, log.stream_name))
            thread.start()
            self.logs = []
            self.appendLogs(log.log_info)

    def updateAndQueueLog(self, currentLog, updates, stream):
        """
        Update an existing log entry with new information and queue it for uploading.

        :param currentLog: Log object to be updated.
        :param updates: Log object containing updates.
        :param stream: Name of the log stream.
        :return: Updated Log object.
        """
        newLogEntities = currentLog.log_info.copy()

        if updates:
            updatedLogEntities = updates.log_info
            for key in newLogEntities:
                if key in updatedLogEntities:
                    newLogEntities[key] = updatedLogEntities[key]

        log = Log(stream, newLogEntities)
        self.queueLog(log)

        return log

    def flushLogs(self, stream_name):
        """
        Flush the current logs buffer, uploading any remaining logs to Azure Monitor.

        :param stream_name: Name of the log stream.
        """
        if self.logs:
            log_batch = self.logs
            thread = threading.Thread(target=self.saveLogs, args=(log_batch, stream_name))
            thread.start()
            self.logs = []



from pyspark.sql import SparkSession
from pyspark.sql.functions import col, lit, concat_ws, sha2, expr
from delta.tables import *
from datetime import *
import pytz
from FabricFramework.FabricLocations import *


class FabricDataInterface:
    """
    Class representing the interface for data operations in the Fabric framework.
    """

    def __init__(self, name="default"):
        """
        Initialize a FabricDataInterface object.

        :param name: Name of the data interface.
        """
        self.name = name
        self.spark = SparkSession.builder.appName("Default_Config").getOrCreate()

    def saveCDCWrapper(self, sourceDataFrame, fabricTableLocation, savetype='cdc-overwrite', primarykey=[],
                       rowhash_columns=[], transactional_watermark_column=None):
        """
        Wrapper method for saving CDC data with support for multiple date partitions.

        :param sourceDataFrame: Source DataFrame to be saved.
        :param fabricTableLocation: FabricTableLocation object representing the target location.
        :param savetype: Save type (e.g., 'overwrite', 'transactions', 'incremental').
        :param primarykey: List of primary key columns.
        :param rowhash_columns: List of columns for row hash.
        :param transactional_watermark_column: Column used for transactional watermark.
        :return: Total record count of the saved data.
        """
        targetRecordCount = 0

        if not sourceDataFrame:
            print("Error occurred during the Fabric CDC save operation: Source Data is empty or does not exist.")
            return targetRecordCount

        if not primarykey and not rowhash_columns:
            print("No Primary Key or Row hash columns have been supplied. Cannot save CDC table.")
            return targetRecordCount

        print(f'CDC {savetype} type save in progress to location {fabricTableLocation.abfss_path()}')

        # Split source data based on distinct 'source_extract_dtm' values if it exists
        list_of_dfs = [sourceDataFrame]
        if "source_extract_dtm" in sourceDataFrame.columns:
            list_of_dfs = []
            source_extract_dtm_distinct = sourceDataFrame.select("source_extract_dtm").distinct().orderBy(
                'source_extract_dtm')
            distinct_value_list = [row["source_extract_dtm"] for row in source_extract_dtm_distinct.collect()]

            for source_extract_dtm in distinct_value_list:
                filtered_df = sourceDataFrame.filter(sourceDataFrame['source_extract_dtm'] == source_extract_dtm)
                list_of_dfs.append(filtered_df)

        for df in list_of_dfs:
            targetRecordCount += self.saveCDCTable(df, fabricTableLocation, savetype, primarykey, rowhash_columns,
                                                   transactional_watermark_column)

        return targetRecordCount

    def saveCDCTable(self, df, fabricTableLocation, savetype='cdc-overwrite', primarykey=[], rowhash_columns=[],
                     transactional_watermark_column=None):
        """
        Save CDC data to a Delta table.

        :param df: DataFrame to be saved.
        :param fabricTableLocation: FabricTableLocation object representing the target location.
        :param savetype: Save type (e.g., 'overwrite', 'transactions', 'incremental').
        :param primarykey: List of primary key columns.
        :param rowhash_columns: List of columns for row hash.
        :param transactional_watermark_column: Column used for transactional watermark.
        :return: Record count of the saved data.
        """
        table_path = fabricTableLocation.abfss_path()

        # Get current datetime and calculate a deducted datetime for record end date
        australiaSydneyTimezone = pytz.timezone("Australia/Sydney")
        current_datetime = datetime.now(australiaSydneyTimezone)
        deducted_datetime = current_datetime - timedelta(seconds=2)
        current_datetime = current_datetime.strftime("%Y-%m-%d %H:%M:%S")
        deducted_datetime = deducted_datetime.strftime("%Y-%m-%d %H:%M:%S")
        highest_datetime = "9999-12-31 00:00:00"

        # Select columns for row hash if not provided
        if not rowhash_columns:
            selected_columns = [col for col in df.columns if col not in primarykey]
        else:
            selected_columns = [col for col in df.columns if col in rowhash_columns and col not in primarykey]

        # Add necessary columns for CDC operations
        df = df.withColumn('key_hash', sha2(concat_ws("_", *primarykey), 256))
        df = df.withColumn('row_hash', sha2(concat_ws("_", *selected_columns), 256))
        df = df.withColumn('Record_Start_Date', expr(f"CAST('{current_datetime}' AS TIMESTAMP)"))
        df = df.withColumn('Record_End_Date', expr(f"CAST('{highest_datetime}' AS TIMESTAMP)"))
        df = df.withColumn('Is_Current', lit(1))
        df = df.withColumn('Is_Deleted', lit(0))
        df = df.withColumn('Last_CDC_Ops', lit("INSERT"))

        # Handle different save types
        if not DeltaTable.isDeltaTable(self.spark, table_path) or savetype == 'cdc-overwrite':
            df.write.mode("overwrite").option("overWriteSchema", "true").format("delta").save(table_path)
            print(f'CDC {savetype} type save to {fabricTableLocation.abfss_path()} complete.')
            return df.count()

        elif savetype == 'cdc-transactions':
            target_df = self.loadLatestDeltaTable(fabricTableLocation)
            print(f'Target Dataframe loaded. It has {target_df.count()} records.')

            watermark_column = df.select(transactional_watermark_column).first()[0]
            source_records_count = target_df.filter(col(transactional_watermark_column) == watermark_column).count()

            print(f'Source data contains {source_records_count} records')

            if source_records_count == 0:
                df.write.mode("append").format("delta").save(table_path)

            print(f'Source has been appended to target. Target now contains {df.count()} records.')

            return df.count()

        elif savetype == 'cdc-incremental':
            target_df = self.loadLatestDeltaTable(fabricTableLocation)
            print(f'Target Dataframe loaded. It has {target_df.count()} records.')

            # Filter the target data for is_current = 1 and is_deleted = 0
            filtered_target_df = target_df.filter((target_df.Is_Current == 1) & (target_df.Is_Deleted == 0))
            newRecords = df.alias("s") \
                .join(filtered_target_df.alias("t"), on=["key_hash"], how="leftanti") \
                .select("s.*")

            print(f'There are {newRecords.count()} new records in the source data.')

            updateRecords = df.alias("s") \
                .join(filtered_target_df.alias("t"), on=["key_hash"], how="inner").where(
                f"t.Is_Current = 1 AND t.Is_Deleted = 0 AND t.row_hash != s.row_hash") \
                .select("s.*")

            print(f'There are {updateRecords.count()} matched (on primary key) records between the target and source data.')

            updateRecords = updateRecords.withColumn("Last_CDC_Ops", lit("UPDATE"))
            updateRecords = updateRecords.withColumn("Record_End_Date",
                                                     expr(f"CAST('{highest_datetime}' AS TIMESTAMP)"))

            mergeRecords = newRecords.union(updateRecords)

            if not mergeRecords.isEmpty():
                print("Merging Source to Target...")

                insertValues = {"key_hash": "s.key_hash", "row_hash": "s.row_hash",
                                "Record_Start_Date": "s.Record_Start_Date", "Record_End_Date": "s.Record_End_Date",
                                "Is_Current": "s.Is_Current", "Is_Deleted": "s.Is_Deleted",
                                "Last_CDC_Ops": "s.Last_CDC_Ops"}

                for c in [i for i in target_df.columns]:
                    insertValues[f"{c}"] = f"s.{c}"

                try:
                    print("Updating Values in Target...")
                    DeltaTable.forPath(self.spark, table_path).alias("t") \
                        .merge(mergeRecords.alias("s"), f"t.key_hash = s.key_hash") \
                        .whenMatchedUpdate(
                        condition=f"t.Is_Current=1 AND t.Is_Deleted = 0 AND t.row_hash != s.row_hash",
                        set={"t.Record_End_Date": expr(f"CAST('{deducted_datetime}' AS TIMESTAMP)"),
                             "t.Is_Current": "0", "t.key_hash": "NULL"}).execute()

                    print("Inserting new values to Target...")
                    DeltaTable.forPath(self.spark, table_path).alias("t") \
                        .merge(mergeRecords.alias("s"), f"t.key_hash = s.key_hash") \
                        .whenNotMatchedInsert(values=insertValues).execute()

                except Exception as e:
                    print(
                        f'Incremental merge operation failed due to Exception Type: {type(e).__name__}, Args: {e.args}')
                    print(e.java_exception)
                else:
                    updated_df = self.loadLatestDeltaTable(fabricTableLocation)
                    print(f'Incremental merge complete. The updated table now contains {updated_df.count()}')
                    return updated_df.count()
            else:
                print("No new or updated records were found in the source data.")
                return 0

        return 0

    def saveDeltaTable(self, sourceDataFrame, fabricTableLocation, savetype='overwrite', primarykey=[]):
        """
        Save data to a Delta table.

        :param sourceDataFrame: Source DataFrame to be saved.
        :param fabricTableLocation: FabricTableLocation object representing the target location.
        :param savetype: Save type (e.g., 'overwrite', 'incremental', 'merge', 'delta').
        :param primarykey: List of primary key columns.
        :return: Record count of the saved data.
        """
        if not sourceDataFrame:
            print("Error occurred during the Fabric Delta Table save operation: Source Data is empty or does not exist.")
            return

        targetRecordCount = 0

        # Check if the target table exists and set the save type accordingly
        if not DeltaTable.isDeltaTable(self.spark, fabricTableLocation.abfss_path()):
            savetype = 'overwrite'
        else:
            # Create the join conditions based on the primary key columns
            if primarykey:
                joinCondition = " AND ".join([f"target.{key} = source.{key}" for key in primarykey])
            else:
                joinCondition = " AND ".join([f"target.{col} = source.{col}" for col in sourceDataFrame.columns])

            target_table = DeltaTable.forPath(self.spark, fabricTableLocation.abfss_path())

        if savetype == 'overwrite':
            print(f'Overwrite Save of table: {fabricTableLocation.abfss_path()}')
            sourceDataFrame.write.mode("overwrite").option("overWriteSchema", "true").format("delta").save(fabricTableLocation.abfss_path())
            targetRecordCount = sourceDataFrame.count()

        elif savetype == 'incremental':
            print(f'Incremental data save to target: {fabricTableLocation.abfss_path()}')
            target_table.alias("target").merge(sourceDataFrame.alias("source"),
                                               joinCondition).whenMatchedUpdateAll().whenNotMatchedInsertAll().execute()
            targetRecordCount = target_table.toDF().count()

        elif savetype == 'merge':
            print(f'Merge data save to target: {fabricTableLocation.abfss_path()}')
            df = sourceDataFrame.withColumn("IsCurrent", lit(True))
            update_set = {"target.IsCurrent": lit(False)}

            target_table.alias("target").merge(df.alias("source"), joinCondition).whenMatchedUpdate(set=update_set).execute()

            matched_rows = df.join(target_table.toDF(), [df[key] == target_table.toDF()[key] for key in primarykey], "left_semi")

            target_table.alias("target").merge(matched_rows.alias("matched"), "false").whenNotMatchedInsertAll().execute()

            targetRecordCount = target_table.toDF().count()

        elif savetype == 'delta':
            print(f'Delta save to target: {fabricTableLocation.abfss_path()}')
            target_table.alias("target").merge(sourceDataFrame.alias("source"),
                                               joinCondition).whenMatchedUpdateAll().whenNotMatchedInsertAll().whenNotMatchedBySourceDelete().execute()
            targetRecordCount = target_table.toDF().count()

        return targetRecordCount

    def loadDeltaLakeTable(self, sourceID, ingestionConfigurationLocation, stage='RAW'):
        """
        Load a Delta Lake table based on source ID and configuration location.

        :param sourceID: Source ID for the table.
        :param ingestionConfigurationLocation: FabricTableLocation object for the ingestion configuration.
        :param stage: Stage of the lakehouse (e.g., 'RAW', 'TRUSTED').
        :return: DataFrame of the loaded Delta Lake table.
        """
        ingestion_manifest_df = self.loadLatestDeltaTable(ingestionConfigurationLocation)
        source = ingestion_manifest_df.filter(col("SourceID") == sourceID)

        if stage == "TRUSTED":
            workspace_id = source.first()['trusted_workspace_id']
            lakehouse_id = source.first()['trusted_lakehouse_id']
            table_name = source.first()['destination_name']
        else:  # defaults to RAW
            workspace_id = source.first()['raw_workspace_id']
            lakehouse_id = source.first()['raw_lakehouse_id']
            table_name = source.first()['destination_name']

        table_location = FabricTableLocation(table_name, FabricLakehouseLocation(stage, workspace_id, lakehouse_id))

        return self.loadLatestDeltaTable(table_location)

    def loadLatestDeltaTable(self, fabricTableLocation):
        """
        Load the latest version of a Delta table.

        :param fabricTableLocation: FabricTableLocation object representing the table location.
        :return: DataFrame of the latest version of the Delta table, or None if the table does not exist.
        """
        table_path = fabricTableLocation.abfss_path()

        if DeltaTable.isDeltaTable(self.spark, table_path):
            latest_version = DeltaTable.forPath(self.spark, table_path).history().select("version").orderBy("version", ascending=False).first()[0]
            latest_data = self.spark.read.format("delta").option("versionAsOf", latest_version).load(table_path)
            return latest_data

        print(f'Could not load Delta table. Location {table_path} does not exist.')
        return None

    def loadIngestionConfiguration(self, fabricFileLocation, fabricTableLocation):
        """
        Load the ingestion configuration from a Delta table or a JSON file.

        :param fabricFileLocation: FabricFileLocation object for the configuration file.
        :param fabricTableLocation: FabricTableLocation object for the configuration table.
        :return: DataFrame of the ingestion configuration.
        """
        ingestion_manifest = self.loadLatestDeltaTable(fabricTableLocation)

        if ingestion_manifest:
            return ingestion_manifest
        else:
            print("Ingestion Configuration Table does not exist, creating from file...")
            ingestion_manifest = self.spark.read.option("multiline", "true").json(fabricFileLocation.abfss_path())
            self.saveDeltaTable(ingestion_manifest, fabricTableLocation)

        return ingestion_manifest
